/*  NoASM compiler for NoSense - The Virtual Machine
    Copyright (C) 2004  Adam Sawicki "Regedit"

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//-------------------------------------------------------------------------
// Name    : Compiler
// Version : 1.0
// Desc    : Module with part 3/3 - compiler
// Author  : Adam Sawicki "Regedit"
// Date    : 2004-06-18
// URL     : mailto:regedit@risp.pl
// URL     : http://www.programex.prv.pl/
//-------------------------------------------------------------------------
#include "stdafx.h"
#include <pxxnl.h>
#include "base.h"
#include "output.h"
#include "Compiler.h"

//-------------------------------------------------------------------------
// Constants

// Compiler commands
const int COMPILER_CMD_COUNT = 55;
const char *COMPILER_CMDS[COMPILER_CMD_COUNT] = {
	"namespace", "end",
	"offset", "offsetdelta",
	"includebin",
	"bootnone", "bootall", "bootend", "boot",
	"data1u", "data2u", "data4u", "data8u",
		"data1i", "data2i", "data4i", "data8i",
		"data4f", "data8f",
		"data0s", "data1s", "data2s", "data4s", "datans",
	"array1u", "array2u", "array4u", "array8u",
		"array1i", "array2i", "array4i", "array8i",
		"array4f", "array8f",
		"array0s", "array1s", "array2s", "array4s", "arrayns",
	"skip1u", "skip2u", "skip4u", "skip8u",
		"skip1i", "skip2i", "skip4i", "skip8i",
		"skip4f", "skip8f",
		"skip0s", "skip1s", "skip2s", "skip4s", "skipns",
	"reserve"
};

//-------------------------------------------------------------------------
// Private data types

typedef std::list<Token*> TokenList2;

class Instruction : public NamedItem {
public:
	enum PARAM_TYPE {
		TYPE_NULL,
		TYPE_1u, TYPE_2u, TYPE_4u, TYPE_8u,
		TYPE_1i, TYPE_2i, TYPE_4i, TYPE_8i,
		TYPE_4f, TYPE_8f,
		TYPE_0s, TYPE_1s, TYPE_2s, TYPE_4s, TYPE_ns,
		TYPE_x
	};
private:
	BYTE category;
	BYTE code;
	PARAM_TYPE params[3];
	int paramCount;
public:
	Instruction(const std::string &a_name, BYTE a_category, BYTE a_code, int a_paramCount, PARAM_TYPE a_p1, PARAM_TYPE a_p2, PARAM_TYPE a_p3);
	BYTE getCategory() { return category; }
	BYTE getCode() { return code; }
	int getParamCount() { return paramCount; }
	PARAM_TYPE getParam(int index) { return params[index]; }
};

Instruction::Instruction(const std::string &a_name, BYTE a_category, BYTE a_code,
	int a_paramCount, PARAM_TYPE a_p1, PARAM_TYPE a_p2, PARAM_TYPE a_p3)
	: NamedItem(a_name)
{
	category = a_category;
	code = a_code;
	paramCount = a_paramCount;
	params[0] = a_p1;
	params[1] = a_p2;
	params[2] = a_p3;
}

typedef UniversalMap<Instruction> InstructionList;

class BinData {
private:
	BYTE *data;
	int size;
public:
	BinData(int a_size) : size(a_size) { data = new BYTE[size]; }
	BinData(int a_size, const void *a_data) : size(a_size) { data = new BYTE[a_size]; memcpy(data, a_data, a_size); }
	~BinData() { delete [] data; }
	void set(const void *a_data) { memcpy(data, a_data, size); }
	void *get() { return data; }
	int getSize() { return size; }
};

class Label : public NamedItem { // name is used as namespace path
private:
	int m_addr;
public:
	Label(const std::string &path, int addr) : NamedItem(path), m_addr(addr) { }
	int getAddr() { return m_addr; }
};

typedef UniversalMap<Label> LabelList;

// Remembered label use that need to be fill out in output file
// (usage was before declaration of this label)
class LabelUse {
private:
	std::string currentPath; // remembered current path from place where label was used
	std::string labelPath;
	DWORD pos; // absolute position in output file
	std::string file;
	int number;
public:
	LabelUse(const std::string &a_currentPath, const std::string &a_labelPath, DWORD a_pos, const std::string &a_file, int a_number) : currentPath(a_currentPath), labelPath(a_labelPath), pos(a_pos), file(a_file), number(a_number) { }
	const std::string &getCurrentPath() const { return currentPath; }
	const std::string &getLabelPath() const { return labelPath; }
	DWORD getPos() const { return pos; }
	const std::string &getFile() { return file; }
	int getNumber() { return number; }
};

typedef UniversalList<LabelUse> LabelUseList;

//-------------------------------------------------------------------------
// Variables

InstructionList g_instructions;
LabelList g_labels;
LabelUseList g_usedLabels;

std::string g_currentPath;
Line *g_currentLine; // currently compiled line (for erroring)

// Boot information to write after compilation
bool g_BootInfoDefined = false;
int g_BootInfo;

//-------------------------------------------------------------------------
// Private functions

void showError_CurrentLine(int code, const std::string &message)
{
	showError(code, message, g_currentLine->getFile(), g_currentLine->getNumber());
}

// Namespace path functions

void nspath_push(std::string *path, const std::string &name)
{
	if (path->empty()) *path = name;
	else *path += '.' + name;
//	std::cout << "NAMESPACE BEGIN - " + s << std::endl;
}

void nspath_cat(std::string *path, const std::string &path2)
{
	nspath_push(path, path2);
}

bool nspath_pop(std::string *path)
{
	// ] path is empty - cannot pop - error
	if (path->empty()) return false;
	// find '.'
	size_t pos = path->find_last_of('.');
	// ] not found - this is single name 
	if (pos == std::string::npos) path->clear();
	// ] found - remove last part
	else path->erase(pos);
//	std::cout << "NAMESPACE END - " + s << std::endl;
	return true;
}

// Load parameter from instruction definitions file (on error: return false)
bool LoadParameter(px::CXNLRead *doc, Instruction::PARAM_TYPE *param)
{
	std::string s;
	if (!doc->ReadElement(&s)) return false;
	if (doc->QuerySeparator() != px::XNL_SEPARATOR_NONE || doc->QueryEOL()) return false;
	
	if (strCiEq(s, "1u")) *param = Instruction::TYPE_1u;
	else if (strCiEq(s, "2u")) *param = Instruction::TYPE_2u;
	else if (strCiEq(s, "4u")) *param = Instruction::TYPE_4u;
	else if (strCiEq(s, "8u")) *param = Instruction::TYPE_8u;
	else if (strCiEq(s, "1i")) *param = Instruction::TYPE_1i;
	else if (strCiEq(s, "2i")) *param = Instruction::TYPE_2i;
	else if (strCiEq(s, "4i")) *param = Instruction::TYPE_4i;
	else if (strCiEq(s, "8i")) *param = Instruction::TYPE_8i;
	else if (strCiEq(s, "4f")) *param = Instruction::TYPE_4f;
	else if (strCiEq(s, "8f")) *param = Instruction::TYPE_8f;
	else if (strCiEq(s, "0s")) *param = Instruction::TYPE_0s;
	else if (strCiEq(s, "1s")) *param = Instruction::TYPE_1s;
	else if (strCiEq(s, "2s")) *param = Instruction::TYPE_2s;
	else if (strCiEq(s, "4s")) *param = Instruction::TYPE_4s;
	else if (strCiEq(s, "ns")) *param = Instruction::TYPE_ns;
	else if (strCiEq(s, "x")) *param = Instruction::TYPE_x;
	else return false;

	return true;
}

// (on error: returns false)
bool LoadInstruction(px::CXNLRead *doc, std::string *extraMsg)
{
	std::string s;

	// name
	std::string name;
	if (!doc->ReadElement(&name)) return false;
	if (doc->QuerySeparator() != px::XNL_SEPARATOR_NONE || doc->QueryEOL()) return false;

/*	// - name already exists
	if (g_instructions.find(name) != 0) {
		*extraMsg = "\""+name+"\" instruction already defined";
		return false;
	}*/

	// category
	if (!doc->ReadElement(&s)) return false;
	if (doc->QuerySeparator() != px::XNL_SEPARATOR_NONE || doc->QueryEOL()) return false;
	BYTE category = static_cast<BYTE>( px::str2int(s) );

	// code
	if (!doc->ReadElement(&s)) return false;
	BYTE code = static_cast<BYTE>( px::str2int(s) );

	// parameters
	int paramCount = 0;
	Instruction::PARAM_TYPE params[3] = { Instruction::TYPE_NULL, Instruction::TYPE_NULL, Instruction::TYPE_NULL };
	if (doc->ReadSectionBegin()) {
		if (doc->QuerySeparator() != px::XNL_SEPARATOR_NONE || doc->QueryEOL()) return false;
		while (!doc->ReadSectionEnd()) {
			if (paramCount >= 3) return false;
			if ( !LoadParameter(doc, &params[paramCount]) ) return false;
			paramCount++;
		}
	}
	if (doc->QuerySeparator() != px::XNL_SEPARATOR_SEMICOLON || !doc->QueryEOL()) return false;

	g_instructions.Add( new Instruction(name, category, code, paramCount, params[0], params[1], params[2]) );
	return true;
}

// Load instruction definitions from file (on error: show message, return false)
bool LoadInstructions()
{
	std::string docFileName = "Instructions.xnl";
	std::string docs, extraMsg;
	bool ok = px::LoadStringFromFile(px::GetModulePath()+docFileName, docs);

	if (ok) {
		px::CXNLRead doc(&docs);
		doc.Begin(px::GetModulePath());
		while (!doc.QueryEOF()) {
			ok = LoadInstruction(&doc, &extraMsg);
			if (!ok) break;
		}
		doc.End();
	}

	if (!ok) showError(ERR_INSTRUCTIONS, "Cannot load instruction definitions from file \""+docFileName+"\"" + (extraMsg.empty() ? "" : " ("+extraMsg+")"));
	return ok;
}

// Returns size of data that should be writen from given tokens as given type
int Tokens2Size(Instruction::PARAM_TYPE Type, TokenList2 *Tokens)
{
	switch (Type)
	{
	case Instruction::TYPE_1u: return 1;
	case Instruction::TYPE_2u: return 2;
	case Instruction::TYPE_4u: return 4;
	case Instruction::TYPE_8u: return 8;
	case Instruction::TYPE_1i: return 1;
	case Instruction::TYPE_2i: return 2;
	case Instruction::TYPE_4i: return 4;
	case Instruction::TYPE_8i: return 8;
	case Instruction::TYPE_4f: return 4;
	case Instruction::TYPE_8f: return 8;
	case Instruction::TYPE_0s: return static_cast<int>(Tokens->front()->str.size())+1;
	case Instruction::TYPE_1s: return static_cast<int>(Tokens->front()->str.size())+1;
	case Instruction::TYPE_2s: return static_cast<int>(Tokens->front()->str.size())+2;
	case Instruction::TYPE_4s: return static_cast<int>(Tokens->front()->str.size())+4;
	case Instruction::TYPE_ns: return static_cast<int>(Tokens->front()->str.size());
	default: return 0; // never
	}
}

// Rewrite given binary file into output file
// (on error: show message)
void IncludeBin(const std::string &FileName)
{
	bool error = false;
	// obtain full file name
	std::string RealFileName = LookForFile(FileName);
	if (RealFileName.empty()) {
		showError_CurrentLine(ERR_FILENOTFOUND, "\""+FileName+"\" file not found");
		error = true;
	}
	// open file for reading
	HANDLE fileh = 0;
	if (!error) {
		fileh = CreateFile(RealFileName.c_str(), GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
		if (fileh == INVALID_HANDLE_VALUE) {
			showError_CurrentLine(ERR_COMPILER_INCLUDEBIN, "\""+RealFileName+"\" cannot open included file");
			error = true;
		}
	}
	// rewrite data
	if (!error) {
		DWORD x;
		BYTE buf[out::BUFFER_SIZE];
		for(;;) {
			if ( !ReadFile(fileh, buf, out::BUFFER_SIZE, &x, 0) ) {
				showError_CurrentLine(ERR_COMPILER_INCLUDEBIN, "\""+FileName+"\" cannot read from included file");
				error = true; break;
			}
			if (x == 0) break;
			if ( !out::writeCode(buf, x) ) {
				// message already shown by out::writeCode
				error = true; break;
			}
		}
	}
	// close file
	if (fileh) CloseHandle(fileh);
}

// Parse label
// On success: return name of read label, move position in line, return true
// On failure: leave position unchanged, return false
bool compileLabel(Line *line, std::string *name)
{
	Line::iterator it = line->getPos();
	if ( tokenParticular(line, Token::IDENTIFIER, name) ) {
		if ( tokenParticular(line, Token::SYMBOL, ":") ) return true;
		else {
			line->setPos(it);
			return false;
		}
	} else return false;
}

// Checks whether label with given name defined in given namespace path is valid
// If yes, return 0
// If no, return label that it conflicts with
// Eg label "a.b" (path = "a", name = "b") is wrong
// when label "a.b" or label "b" already exists.
Label *CheckLabel(const std::string &Path, const std::string &Name)
{
	std::string Path2 = Path;
	std::string FullPath;
	Label *label;
	for(;;) {
		FullPath = Path2;
		nspath_push(&FullPath, Name);
		label = g_labels.find(FullPath);
		if (label) return label;
		if (!nspath_pop(&Path2)) break;
	}
	return 0;
}

// Add to global label list label with given name
// Full path is constructed from given name and the current namespace path
// Label address is the current address
bool addLabel(const std::string &name)
{
	std::string path = g_currentPath;
	nspath_push(&path, name);
	Label *label = g_labels.find(path);
	if (label == 0) {
		label = CheckLabel(g_currentPath, name);
		if (label == 0) {
			g_labels.Add( new Label(path, out::getAddress()) );
			stats::labelDefined++;
			return true;
		}
		else {
			showError_CurrentLine(ERR_COMPILER_LBLREDEF, "\""+path+"\" label conflicts with \""+label->getName()+"\"");
			return false;
		}
	}
	else {
		showError_CurrentLine(ERR_COMPILER_LBLREDEF, "\""+path+"\" label already defined");
		return false;
	}
}

// Compile a parameter from given line (on error: show message, return false)
// Return its:
// - addressing mode (forced by first addressing mode symbol or deduced from kind of token)
// - tokens (single number, single string or a chain of identifier.identifier.identifier...
//   as a label) - if function succeedes, list has always at least one element
// - optional label name (if no label, this name is empty)
bool compileParameter(Line *line, BYTE *addrMode, TokenList2 *tokens, std::string *label)
{
	// 1. optional comma
	tokenParticular(line, Token::SYMBOL, ",");

	// 2. label
	if (compileLabel(line, label)) { }
	else label->clear();

	// 3. addressing mode
	*addrMode = 0;
	for (;;) {
		// only first is taken into account
		if ( tokenParticular(line, Token::SYMBOL, "&") ) {
			if (*addrMode == 0) *addrMode = 1;
		}
		else if ( tokenParticular(line, Token::SYMBOL, "^") ) {
			if (*addrMode == 0) *addrMode = 2;
		}
		else if ( tokenParticular(line, Token::SYMBOL, "*") ) {
			if (*addrMode == 0) *addrMode = 3;
		}
		else break;
	}

	// 4. parameter tokens
	if (line->queryEnd()) {
		Expected(ERR_EXP_PARAM, line);
		return false;
	}
	tokens->clear();
	tokens->push_back( line->at() );
	line->setPosNext();
	if (tokens->front()->type == Token::IDENTIFIER) {
		while (!line->queryEnd()) {
			if ( line->at()->type == Token::SYMBOL && line->at()->str == "." ) {
				tokens->push_back( line->at() );
				line->setPosNext();
			}
			else break;
			if ( line->at()->type == Token::IDENTIFIER ) {
				tokens->push_back( line->at() );
				line->setPosNext();
			}
			else {
				Expected(ERR_EXP_ID, line);
				return false;
			}
		}
	}

	// if no addressing mode is forced - we must deduce it from parameter token
	if (*addrMode == 0) {
		// identifier means label - default addressing mode is 2
		if ( ( tokens->front() )->type == Token::IDENTIFIER ) *addrMode = 2;
		// no identifier (number or string) means direct value - default addressing mode is 1
		else *addrMode = 1;
	}

	return true;
}

bool char2digit(char c, int base, int *digit)
{
	if (c >= '0' && c <= '9') *digit = static_cast<int>(c - '0');
	else if (c >= 'a' && c <= 'z') *digit = static_cast<int>(c - 'a') + 10;
	else if (c >= 'A' && c <= 'Z') *digit = static_cast<int>(c - 'A') + 10;
	else return false;
	return (*digit < base);
}

bool parseNumber_float(Token *token, float *v)
{
	*v = px::str2float(token->str);
	return true;
}

bool parseNumber_double(Token *token, double *v)
{
	*v = px::str2double(token->str);
	return true;
}

// Parse given number token as the unsigned integer value and return the value
// (on error: return false)
template <typename T>
bool parseNumber_u(Token *token, T *v)
{
	*v = 0;
	std::string s = token->str;
	int base = 10, digit;
	// if string is empty (it should never be)
	if (s.size() == 0) return false;
	// '+' or '-' at the begin
	if (s[0] == '+' || s[0] == '-') {
		s.erase(0, 1);
		if (s.size() == 0) return false;
	}
	// '0x' at the begin
	if (s.size() >= 2 && px::LowerCase(s.substr(0, 2)) == "0x") {
		base = 16;
		s.erase(0, 2);
		if (s.size() == 0) return false;
	}
	// 'b' or 'd' or 'o' or 'h' at the end
	else {
		char c = s[s.size()-1];
		if (c == 'b' || c == 'B') {
			base = 2; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
		else if (c == 'o' || c == 'O') {
			base = 8; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
		else if (c == 'd' || c == 'D') {
			base = 10; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
		else if (c == 'h' || c == 'H') {
			base = 16; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
	}
	while (!s.empty()) {
		if (!char2digit(s[0], base, &digit)) return false;
		s.erase(0, 1);
		*v = *v * base + digit;
	}
	return true;
}

// Parse given number token as the signed integer value and return the value
// (on error: return false)
template <typename T>
bool parseNumber_i(Token *token, T *v)
{
	*v = 0;
	std::string s = token->str;
	int base = 10, digit, sign = 1;
	// if string is empty (it should never be)
	if (s.size() == 0) return false;
	// '+' or '-' at the begin
	if (s[0] == '+' || s[0] == '-') {
		if (s[0] == '-') sign = -1;
		s.erase(0, 1);
		if (s.size() == 0) return false;
	}
	// '0x' at the begin
	if (s.size() >= 2 && px::LowerCase(s.substr(0, 2)) == "0x") {
		base = 16;
		s.erase(0, 2);
		if (s.size() == 0) return false;
	}
	// 'b' or 'd' or 'o' or 'h' at the end
	else {
		char c = s[s.size()-1];
		if (c == 'b' || c == 'B') {
			base = 2; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
		else if (c == 'o' || c == 'O') {
			base = 8; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
		else if (c == 'd' || c == 'D') {
			base = 10; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
		else if (c == 'h' || c == 'H') {
			base = 16; s.erase(s.size()-1, 1); if (s.size() == 0) return false;
		}
	}
	while (!s.empty()) {
		if (!char2digit(s[0], base, &digit)) return false;
		s.erase(0, 1);
		*v = *v * base + digit * sign;
	}
	return true;
}

// Convert list of tokens to namespace path
// (on error: show message, return anything)
std::string TokenList2LabelPath(TokenList2 *tokens)
{
	std::string s;
	bool ok = true;
	TokenList2::iterator it = tokens->begin();

	if (ok) if ( it == tokens->end() ) ok = false;
	if (ok) if ( (*it)->type != Token::IDENTIFIER ) ok = false;
	if (ok) { s += (*it)->str; ++it; }

	if (ok) {
		while ( it != tokens->end() ) {
			// '.'
			if ( (*it)->type != Token::SYMBOL || (*it)->str != "." ) { ok = false; break; }
			++it;
			// next identifier
			if ( it == tokens->end() ) { ok = false; break; }
			if ( (*it)->type != Token::IDENTIFIER ) { ok = false; break; }
			s += '.';
			s += (*it)->str;
			++it;
		}
	}
	if (!ok) showError_CurrentLine(ERR_INTERNAL, "Internal error: cannot convert token list to namespace path");
	return s;
}

// Return label address
// Label is given by current path and label path
// - If label is defined, return its address, return true
// - If not, return false
bool findLabel(const std::string &a_currentPath, const std::string &labelPath, int *addr)
{
	// find existing label that matches
	std::string path, currentPath = a_currentPath;
	Label *label;
	for(;;) {
		path = currentPath;
		nspath_cat(&path, labelPath);
		label = g_labels.find(path);
		// label found
		if (label) {
			*addr = label->getAddr();
			return true;
		}
		// try path one level up
		if ( !nspath_pop(&currentPath) ) break;
	}
	// label not found
	return false;
}

// Use label - return its address (on error: show message)
// Label is given by token list and current namespace path
// - If label is defined, return its address
// - If not, return anything and save information that it should be completed later
int useLabel(TokenList2 *tokens)
{
	stats::labelUsed++;
	int addr;
	std::string labelPath = TokenList2LabelPath(tokens);
	if ( findLabel(g_currentPath, labelPath, &addr) )
		return addr;
	else {
		g_usedLabels.Add( new LabelUse(g_currentPath, labelPath, out::getPos(), g_currentLine->getFile(), g_currentLine->getNumber()) );
		return 0;
	}
}

// #### parseValue_* ####
// Parse token list to a value of particular type and return this value
// (on error: show error message, return false)

template<typename T>
bool parseValue_u(TokenList2 *tokens, T *value)
{
	// single character in a string
	if (tokens->front()->type == Token::STRING && tokens->front()->str.size() == 1) {
		*value = static_cast<T>(tokens->front()->str[0]);
		return true;
	}
	// number
	if (tokens->front()->type == Token::NUMBER) {
		// integer number
		if ( parseNumber_u<T>(tokens->front(), value) )
			return true;
		// floating point number
		double v;
		if ( parseNumber_double(tokens->front(), &v) ) {
			*value = static_cast<T>(v);
			return true;
		}
	}
	// something other (not a number or cannot parse number as int. or float.)
	Expected(ERR_EXP_NR, g_currentLine);
	return false;
}

template<typename T>
bool parseValue_i(TokenList2 *tokens, T *value)
{
	// single character in a string
	if (tokens->front()->type == Token::STRING && tokens->front()->str.size() == 1) {
		*value = static_cast<T>(tokens->front()->str[0]);
		return true;
	}
	// number
	if (tokens->front()->type == Token::NUMBER) {
		// integer number
		if ( parseNumber_i<T>(tokens->front(), value) )
			return true;
		// floating point number
		double v;
		if ( parseNumber_double(tokens->front(), &v) ) {
			*value = static_cast<T>(v);
			return true;
		}
	}
	// something other (not a number or cannot parse number as int. or float.)
	Expected(ERR_EXP_NR, g_currentLine);
	return false;
}

bool parseValue_i_4(TokenList2 *tokens, signed __int32 *value)
{
	typedef signed __int32 T;
	// single character in a string
	if (tokens->front()->type == Token::STRING && tokens->front()->str.size() == 1) {
		*value = static_cast<T>(tokens->front()->str[0]);
		return true;
	}
	// label
	if (tokens->front()->type == Token::IDENTIFIER) {
		*value = useLabel(tokens);
		return true;
	}
	// number
	if (tokens->front()->type == Token::NUMBER) {
		// integer number
		if ( parseNumber_i<T>(tokens->front(), value) )
			return true;
		// floating point number
		double v;
		if ( parseNumber_double(tokens->front(), &v) ) {
			*value = static_cast<T>(v);
			return true;
		}
	}
	// something other (not a number or cannot parse number as int. or float.)
	Expected(ERR_EXP_NR, g_currentLine);
	return false;
}

bool parseValue_f_4(TokenList2 *tokens, float *value)
{
	typedef float T;
	// single character in a string
	if (tokens->front()->type == Token::STRING && tokens->front()->str.size() == 1) {
		*value = static_cast<T>(tokens->front()->str[0]);
		return true;
	}
	// number
	if (tokens->front()->type == Token::NUMBER) {
		__int64 v;
		// integer number
		if ( parseNumber_i<__int64>(tokens->front(), &v) ) {
			*value = static_cast<T>(v);
			return true;
		}
		// floating point number
		else if ( parseNumber_float(tokens->front(), value) )
			return true;
	}
	// something other (not a number or cannot parse number as int. or float.)
	Expected(ERR_EXP_NR, g_currentLine);
	return false;
}

bool parseValue_f_8(TokenList2 *tokens, double *value)
{
	typedef float T;
	// single character in a string
	if (tokens->front()->type == Token::STRING && tokens->front()->str.size() == 1) {
		*value = static_cast<T>(tokens->front()->str[0]);
		return true;
	}
	// number
	if (tokens->front()->type == Token::NUMBER) {
		__int64 v;
		// integer number
		if ( parseNumber_i<__int64>(tokens->front(), &v) ) {
			*value = static_cast<T>(v);
			return true;
		}
		// floating point number
		else if ( parseNumber_double(tokens->front(), value) )
			return true;
	}
	// something other (not a number or cannot parse number as int. or float.)
	Expected(ERR_EXP_NR, g_currentLine);
	return false;
}

bool parseValue_s(TokenList2 *tokens, std::string *value)
{
	if (tokens->front()->type == Token::STRING) {
		*value = tokens->front()->str;
		return true;
	}
	Expected(ERR_EXP_STR, g_currentLine);
	return false;
}

// #### writeParameter_* ####
// Write into output file value from given tokens as a parameter
// (on error: write message, return false)

template <typename T>
bool writeParameter_u(TokenList2 *tokens)
{
	T v;
	if ( parseValue_u(tokens, &v) )
		return out::writeCode( &v, sizeof(v) );
	else return false;
}

template <typename T>
bool writeParameter_i(TokenList2 *tokens)
{
	T v;
	if ( parseValue_i(tokens, &v) )
		return out::writeCode( &v, sizeof(v) );
	else return false;
}

// Specialization of the function above
bool writeParameter_i_4(TokenList2 *tokens)
{
	typedef signed __int32 T;
	T v;
	if ( parseValue_i_4(tokens, &v) )
		return out::writeCode( &v, sizeof(v) );
	else return false;
}

bool writeParameter_4f(TokenList2 *tokens)
{
	float v;
	if ( parseValue_f_4(tokens, &v) )
		return out::writeCode( &v, sizeof(v) );
	else return false;
}

bool writeParameter_8f(TokenList2 *tokens)
{
	double v;
	if ( parseValue_f_8(tokens, &v) )
		return out::writeCode( &v, sizeof(v) );
	else return false;
}

bool writeParameter_1u(TokenList2 *tokens) { return writeParameter_u<unsigned __int8>(tokens); }
bool writeParameter_2u(TokenList2 *tokens) { return writeParameter_u<unsigned __int16>(tokens); }
bool writeParameter_4u(TokenList2 *tokens) { return writeParameter_u<unsigned __int32>(tokens); }
bool writeParameter_8u(TokenList2 *tokens) { return writeParameter_u<unsigned __int64>(tokens); }

bool writeParameter_1i(TokenList2 *tokens) { return writeParameter_i<signed __int8>(tokens); }
bool writeParameter_2i(TokenList2 *tokens) { return writeParameter_i<signed __int16>(tokens); }
bool writeParameter_4i(TokenList2 *tokens) { return writeParameter_i_4(tokens); }
bool writeParameter_8i(TokenList2 *tokens) { return writeParameter_i<signed __int64>(tokens); }

bool writeParameter_0s(TokenList2 *tokens)
{
	std::string v;
	if ( parseValue_s(tokens, &v) )
		return out::writeCode( v.c_str(), static_cast<int>(v.size()+1) );
	return false;
}

bool writeParameter_1s(TokenList2 *tokens)
{
	std::string v;
	if ( parseValue_s(tokens, &v) ) {
		typedef unsigned __int8 T;
		T size = static_cast<T>( v.size() );
		if (! out::writeCode(&size, sizeof(size)) ) return false;
		return out::writeCode( v.c_str(), static_cast<int>(v.size()) );
	}
	return false;
}

bool writeParameter_2s(TokenList2 *tokens)
{
	std::string v;
	if ( parseValue_s(tokens, &v) ) {
		typedef unsigned __int16 T;
		T size = static_cast<T>( v.size() );
		if (! out::writeCode(&size, sizeof(size)) ) return false;
		return out::writeCode( v.c_str(), static_cast<int>(v.size()) );
	}
	return false;
}

bool writeParameter_4s(TokenList2 *tokens)
{
	std::string v;
	if ( parseValue_s(tokens, &v) ) {
		typedef unsigned __int32 T;
		T size = static_cast<T>( v.size() );
		if (! out::writeCode(&size, sizeof(size)) ) return false;
		return out::writeCode( v.c_str(), static_cast<int>(v.size()) );
	}
	return false;
}

bool writeParameter_ns(TokenList2 *tokens)
{
	std::string v;
	if ( parseValue_s(tokens, &v) )
		return out::writeCode( v.c_str(), static_cast<int>(v.size()) );
	return false;
}

// Write given tokens as a parameter with given type and addressing mode
// (on error: show message, return false)
bool writeParameter(Instruction::PARAM_TYPE type, BYTE addrMode, TokenList2 *tokens)
{
	if (addrMode == 1) {
		switch (type)
		{
		case Instruction::TYPE_1u: return writeParameter_1u(tokens);
		case Instruction::TYPE_2u: return writeParameter_2u(tokens);
		case Instruction::TYPE_4u: return writeParameter_4u(tokens);
		case Instruction::TYPE_8u: return writeParameter_8u(tokens);
		case Instruction::TYPE_1i: return writeParameter_1i(tokens);
		case Instruction::TYPE_2i: return writeParameter_2i(tokens);
		case Instruction::TYPE_4i: return writeParameter_4i(tokens);
		case Instruction::TYPE_8i: return writeParameter_8i(tokens);
		case Instruction::TYPE_4f: return writeParameter_4f(tokens);
		case Instruction::TYPE_8f: return writeParameter_8f(tokens);
		case Instruction::TYPE_0s: return writeParameter_0s(tokens);
		case Instruction::TYPE_1s: return writeParameter_1s(tokens);
		case Instruction::TYPE_2s: return writeParameter_2s(tokens);
		case Instruction::TYPE_4s: return writeParameter_4s(tokens);
		case Instruction::TYPE_ns: return writeParameter_ns(tokens);
		default: /* Instruction::TYPE_x */
			showError_CurrentLine(ERR_COMPILER_XTYPE, "Cannot write \"x\" type parameter as a direct value (addressing mode 1)");
			return false;
		}
	}
	else if (addrMode == 2 || addrMode == 3) {
		return writeParameter_4i(tokens);
	}
	else {
		showError_CurrentLine(ERR_INTERNAL, "Internal error: invalid addressing mode in writeParameter()");
		return false;
	}
}

// Compile particular compiler command from given line
// (command name is already read from the line, we must only read parameters)

bool compileNamespaceBegin(Line *line)
{
	std::string name;
	// read namespace name
	if (tokenParticular(line, Token::IDENTIFIER, &name)) { }
	// no name - unnamed namespace
	else generateRandomString(&name);
	// go into new (maybe sub-) namespace
	nspath_push(&g_currentPath, name);
	// shoud be end of line
	if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);
	return true;
}

bool compileNamespaceEnd(Line *line)
{
	// shoud be end of line
	if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);
	// go into from the most internal namespace
	if (!nspath_pop(&g_currentPath))
		showError_CurrentLine(ERR_COMPILER_NSEND, "No namespace to end");
	return true;
}

bool compileOffset(Line *line, bool relative)
{
	// read address
	BYTE addrMode; TokenList2 tokens; std::string label;
	if ( compileParameter(line, &addrMode, &tokens, &label) ) {
		if (addrMode != 1)
			showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
		else if ( !label.empty() )
			showError_CurrentLine(ERR_COMPILER_NOLABEL, ERRMSG_COMPILER_NOLABEL);
		else {
			int addr;
			if ( parseValue_i_4(&tokens, &addr) ) {
				if (relative)
					addr += out::getAddress();
				out::setAddress(addr);
			}
		}
		// shoud be end of line
		if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);
	}
	return true;
}

bool compileIncludeBin(Line *line)
{
	// read file name
	BYTE addrMode; TokenList2 tokens; std::string label;
	if ( compileParameter(line, &addrMode, &tokens, &label) ) {
		if (addrMode != 1)
			showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
		else if ( !label.empty() )
			showError_CurrentLine(ERR_COMPILER_NOLABEL, ERRMSG_COMPILER_NOLABEL);
		else {
			std::string FileName;
			if ( parseValue_s(&tokens, &FileName) ) {
				IncludeBin(FileName);
			}
		}
		// shoud be end of line
		if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);
	}
	return true;
}

enum BOOT_INFO_TYPE {
	BOOT_GIVEN, // given particular boot info
	BOOT_CURRENT, // current position
	BOOT_PARAM // info given in instruction parameter
};

bool compileBoot(Line *line, BOOT_INFO_TYPE BootInfoType, int info = 0)
{
	// get info
	bool ok = true;
	int NewBootInfo;
	if (BootInfoType == BOOT_GIVEN) {
		NewBootInfo = info;
		// shoud be end of line
		if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);
	}
	else if (BootInfoType == BOOT_CURRENT) {
		NewBootInfo = out::getCodeSize();
		// shoud be end of line
		if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);
	}
	else { // BootInfoType == BOOT_PARAM
		// read info
		BYTE addrMode; TokenList2 tokens; std::string label;
		if ( compileParameter(line, &addrMode, &tokens, &label) ) {
			if (addrMode != 1) showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
			else if ( !label.empty() ) showError_CurrentLine(ERR_COMPILER_NOLABEL, ERRMSG_COMPILER_NOLABEL);
			else {
				if ( parseValue_i_4(&tokens, &NewBootInfo) ) { }
				else ok = false;
			}
			// shoud be end of line
			if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);
		} else ok = false;
	}
	// remember info
	if (ok) {
		if (g_BootInfoDefined)
			showError_CurrentLine(ERR_COMPILER_BOOTINFOREDEF, ERRMSG_COMPILER_BOOTINFOREDEF);
		else {
			g_BootInfo = NewBootInfo;
			g_BootInfoDefined = true;
		}
	}
	return true;
}

bool compileData(Line *line, Instruction::PARAM_TYPE type)
{
	BYTE AddrMode;
	TokenList2 Tokens;
	std::string Label;
	while ( !line->queryEnd() ) {
		if ( !compileParameter(line, &AddrMode, &Tokens, &Label) ) return false;
		if (AddrMode != 1) {
			showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
			return false;
		}
		if ( !Label.empty() ) {
			if (! addLabel(Label) ) return false;
		}
		if (! writeParameter(type, 1, &Tokens) ) return false;
	}
	return true;
}

bool compileArray(Line *line, Instruction::PARAM_TYPE type)
{
	BYTE AddrMode;
	TokenList2 Tokens;
	std::string Label;

	// Element count
	int Count;
	if (compileParameter(line, &AddrMode, &Tokens, &Label)) {
		if (AddrMode != 1) {
			showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
			return false;
		}
		if (!Label.empty()) {
			showError_CurrentLine(ERR_COMPILER_NOLABEL, ERRMSG_COMPILER_NOLABEL);
			return false;
		}
		if ( parseValue_i_4(&Tokens, &Count) ) { }
		else return false;
	}
	else return false;

	// Element
	if (compileParameter(line, &AddrMode, &Tokens, &Label)) {
		if (AddrMode != 1) {
			showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
			return false;
		}
		if (!Label.empty()) {
			showError_CurrentLine(ERR_COMPILER_NOLABEL, ERRMSG_COMPILER_NOLABEL);
			return false;
		}
		for (int i = 0; i < Count; i++) {
			if ( !writeParameter(type, 1, &Tokens) ) return false;
		}
	}
	else return false;

	// Should be end of line
	if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);

	return true;
}

bool compileSkip(Line *line, Instruction::PARAM_TYPE type)
{
	typedef std::list<TokenList2> TokenList2List;

	// Read data
	BYTE AddrMode;
	TokenList2List Tokens;
	StringList Labels;
	TokenList2List::iterator tit;
	StringList::iterator lit;
	while ( !line->queryEnd() ) {
		Tokens.push_back(TokenList2());
		Labels.push_back("");
		if ( !compileParameter(line, &AddrMode, &Tokens.back(), &Labels.back()) ) return false;
		if (AddrMode != 1) {
			showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
			return false;
		}
	}

	// Compute data size
	int Size = 0;
	for (tit = Tokens.begin(); tit != Tokens.end(); ++tit)
		Size += Tokens2Size(type, &(*tit));

	// write instruction
	// - write category and addressing modes
	BYTE b = (0 << 6) | (1 << 4) | (0 << 2) | (0 << 0);
	if (!out::writeCode(&b, 1)) return false;
	// - write instruction code and data size
	if (Size < 256) {
		b = 200; // SKIP1
		if (!out::writeCode(&b, 1)) return false;
		BYTE Size2 = static_cast<BYTE>(Size);
		if (!out::writeCode(&Size2, sizeof(Size2))) return false;
	}
	else if (Size < 256*256) {
		b = 201; // SKIP2
		if (!out::writeCode(&b, 1)) return false;
		WORD Size2 = static_cast<WORD>(Size);
		if (!out::writeCode(&Size2, sizeof(Size2))) return false;
	}
	else {
		b = 202; // SKIP4
		if (!out::writeCode(&b, 1)) return false;
		DWORD Size2 = static_cast<DWORD>(Size);
		if (!out::writeCode(&Size2, sizeof(Size2))) return false;
	}
	// - write data to skip
	for (tit = Tokens.begin(), lit = Labels.begin(); tit != Tokens.end(); ++tit, ++lit) {
		if (!lit->empty()) {
			if (! addLabel(*lit) ) return false;
		}
		if (! writeParameter(type, 1, &(*tit)) ) return false;
	}
	stats::instructions++;
	return true;
}

bool compileReserve(Line *line)
{
	BYTE AddrMode;
	TokenList2 Tokens;
	std::string Label;

	// Read size
	int Size;
	if (compileParameter(line, &AddrMode, &Tokens, &Label)) {
		if (AddrMode != 1) {
			showError_CurrentLine(ERR_COMPILER_ADDRMODE1, ERRMSG_COMPILER_ADDRMODE1);
			return false;
		}
		if (!Label.empty()) {
			showError_CurrentLine(ERR_COMPILER_NOLABEL, ERRMSG_COMPILER_NOLABEL);
			return false;
		}
		if ( parseValue_i_4(&Tokens, &Size) ) { }
		else return false;
	}
	else return false;

	// Should be end of line
	if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);

	out::reserve(Size);

	return true;
}

// Compile given instruction from given line
// (instruction name is already read from the line, we must only read parameters)
bool compileInstruction(Line *line, Instruction *instruction)
{
	int i;
	BYTE paramAddrModes[3] = { 0, 0, 0 }; // must be initialized
	TokenList2 paramTokens[3];
	std::string paramLabels[3]; // empty string means no label
	// read parameters
	for (i = 0; i < instruction->getParamCount(); i++) {
		if (! compileParameter(line, &paramAddrModes[i], &paramTokens[i], &paramLabels[i]) ) return false;
	}
	// shoud be end of line
	if (!line->queryEnd()) Expected(ERR_EXP_EOL, line);

	// write instruction
	// - write category and addressing modes
	BYTE b = (instruction->getCategory() << 6) | (paramAddrModes[0] << 4) | (paramAddrModes[1] << 2) | (paramAddrModes[2] << 0);
	if (!out::writeCode(&b, 1)) return false;
	// - write instruction code
	b = instruction->getCode();
	if (!out::writeCode(&b, 1)) return false;
	// - write parameters
	for (i = 0; i < instruction->getParamCount(); i++) {
		if (! paramLabels[i].empty() ) {
			if (! addLabel(paramLabels[i]) ) return false;
		}
		if (! writeParameter(instruction->getParam(i), paramAddrModes[i], &paramTokens[i]) ) return false;
	}
	stats::instructions++;
	return true;
}

// Compile current source line (on error: show message, return false)
bool compileLine()
{
	g_currentLine = g_source.at();

	Line *line = g_source.at();
	line->setPosFirst();

	std::string name;
	Instruction *instruction;

	// label at the begin
	if ( compileLabel(line, &name) )
		addLabel(name);

	// line is empty
	if (line->queryEnd()) {
		return true;
	}
	else if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		// compiler commands
		if ( strCiEq(name, "namespace") ) {
			if (!compileNamespaceBegin(line)) return false; return true;
		}
		else if ( strCiEq(name, "end") ) {
			if (!compileNamespaceEnd(line)) return false; return true;
		}
		else if ( strCiEq(name, "offset") ) {
			if (!compileOffset(line, false)) return false; return true;
		}
		else if ( strCiEq(name, "offsetdelta") ) {
			if (!compileOffset(line, true)) return false; return true;
		}
		else if ( strCiEq(name, "includebin") ) {
			if (!compileIncludeBin(line)) return false; return true;
		}
		else if ( strCiEq(name, "bootnone") ) {
			if (!compileBoot(line, BOOT_GIVEN, 0)) return false; return true;
		}
		else if ( strCiEq(name, "bootall") ) {
			if (!compileBoot(line, BOOT_GIVEN, -1)) return false; return true;
		}
		else if ( strCiEq(name, "bootend") ) {
			if (!compileBoot(line, BOOT_CURRENT)) return false; return true;
		}
		else if ( strCiEq(name, "boot") ) {
			if (!compileBoot(line, BOOT_PARAM)) return false; return true;
		}
		// - DATA[type]
		else if ( strCiEq(name, "data1u") ) { if (!compileData(line, Instruction::TYPE_1u)) return false; return true; }
		else if ( strCiEq(name, "data2u") ) { if (!compileData(line, Instruction::TYPE_2u)) return false; return true; }
		else if ( strCiEq(name, "data4u") ) { if (!compileData(line, Instruction::TYPE_4u)) return false; return true; }
		else if ( strCiEq(name, "data8u") ) { if (!compileData(line, Instruction::TYPE_8u)) return false; return true; }
		else if ( strCiEq(name, "data1i") ) { if (!compileData(line, Instruction::TYPE_1i)) return false; return true; }
		else if ( strCiEq(name, "data2i") ) { if (!compileData(line, Instruction::TYPE_2i)) return false; return true; }
		else if ( strCiEq(name, "data4i") ) { if (!compileData(line, Instruction::TYPE_4i)) return false; return true; }
		else if ( strCiEq(name, "data8i") ) { if (!compileData(line, Instruction::TYPE_8i)) return false; return true; }
		else if ( strCiEq(name, "data4f") ) { if (!compileData(line, Instruction::TYPE_4f)) return false; return true; }
		else if ( strCiEq(name, "data8f") ) { if (!compileData(line, Instruction::TYPE_8f)) return false; return true; }
		else if ( strCiEq(name, "data0s") ) { if (!compileData(line, Instruction::TYPE_0s)) return false; return true; }
		else if ( strCiEq(name, "data1s") ) { if (!compileData(line, Instruction::TYPE_1s)) return false; return true; }
		else if ( strCiEq(name, "data2s") ) { if (!compileData(line, Instruction::TYPE_2s)) return false; return true; }
		else if ( strCiEq(name, "data4s") ) { if (!compileData(line, Instruction::TYPE_4s)) return false; return true; }
		else if ( strCiEq(name, "datans") ) { if (!compileData(line, Instruction::TYPE_ns)) return false; return true; }
		// - ARRAY[type]
		else if ( strCiEq(name, "array1u") ) { if (!compileArray(line, Instruction::TYPE_1u)) return false; return true; }
		else if ( strCiEq(name, "array2u") ) { if (!compileArray(line, Instruction::TYPE_2u)) return false; return true; }
		else if ( strCiEq(name, "array4u") ) { if (!compileArray(line, Instruction::TYPE_4u)) return false; return true; }
		else if ( strCiEq(name, "array8u") ) { if (!compileArray(line, Instruction::TYPE_8u)) return false; return true; }
		else if ( strCiEq(name, "array1i") ) { if (!compileArray(line, Instruction::TYPE_1i)) return false; return true; }
		else if ( strCiEq(name, "array2i") ) { if (!compileArray(line, Instruction::TYPE_2i)) return false; return true; }
		else if ( strCiEq(name, "array4i") ) { if (!compileArray(line, Instruction::TYPE_4i)) return false; return true; }
		else if ( strCiEq(name, "array8i") ) { if (!compileArray(line, Instruction::TYPE_8i)) return false; return true; }
		else if ( strCiEq(name, "array4f") ) { if (!compileArray(line, Instruction::TYPE_4f)) return false; return true; }
		else if ( strCiEq(name, "array8f") ) { if (!compileArray(line, Instruction::TYPE_8f)) return false; return true; }
		else if ( strCiEq(name, "array0s") ) { if (!compileArray(line, Instruction::TYPE_0s)) return false; return true; }
		else if ( strCiEq(name, "array1s") ) { if (!compileArray(line, Instruction::TYPE_1s)) return false; return true; }
		else if ( strCiEq(name, "array2s") ) { if (!compileArray(line, Instruction::TYPE_2s)) return false; return true; }
		else if ( strCiEq(name, "array4s") ) { if (!compileArray(line, Instruction::TYPE_4s)) return false; return true; }
		else if ( strCiEq(name, "arrayns") ) { if (!compileArray(line, Instruction::TYPE_ns)) return false; return true; }
		// - SKIP[type]
		else if ( strCiEq(name, "skip1u") ) { if (!compileSkip(line, Instruction::TYPE_1u)) return false; return true; }
		else if ( strCiEq(name, "skip2u") ) { if (!compileSkip(line, Instruction::TYPE_2u)) return false; return true; }
		else if ( strCiEq(name, "skip4u") ) { if (!compileSkip(line, Instruction::TYPE_4u)) return false; return true; }
		else if ( strCiEq(name, "skip8u") ) { if (!compileSkip(line, Instruction::TYPE_8u)) return false; return true; }
		else if ( strCiEq(name, "skip1i") ) { if (!compileSkip(line, Instruction::TYPE_1i)) return false; return true; }
		else if ( strCiEq(name, "skip2i") ) { if (!compileSkip(line, Instruction::TYPE_2i)) return false; return true; }
		else if ( strCiEq(name, "skip4i") ) { if (!compileSkip(line, Instruction::TYPE_4i)) return false; return true; }
		else if ( strCiEq(name, "skip8i") ) { if (!compileSkip(line, Instruction::TYPE_8i)) return false; return true; }
		else if ( strCiEq(name, "skip4f") ) { if (!compileSkip(line, Instruction::TYPE_4f)) return false; return true; }
		else if ( strCiEq(name, "skip8f") ) { if (!compileSkip(line, Instruction::TYPE_8f)) return false; return true; }
		else if ( strCiEq(name, "skip0s") ) { if (!compileSkip(line, Instruction::TYPE_0s)) return false; return true; }
		else if ( strCiEq(name, "skip1s") ) { if (!compileSkip(line, Instruction::TYPE_1s)) return false; return true; }
		else if ( strCiEq(name, "skip2s") ) { if (!compileSkip(line, Instruction::TYPE_2s)) return false; return true; }
		else if ( strCiEq(name, "skip4s") ) { if (!compileSkip(line, Instruction::TYPE_4s)) return false; return true; }
		else if ( strCiEq(name, "skipns") ) { if (!compileSkip(line, Instruction::TYPE_ns)) return false; return true; }
		// - RESERVE
		else if ( strCiEq(name, "reserve") ) { if (!compileReserve(line)) return false; return true; }
		// instruction
		else if ( (instruction = g_instructions.find(name)) != 0 ) {
			if (!compileInstruction(line, instruction)) return false;
			return true;
		}
		else {
			showError_CurrentLine(ERR_COMPILER_UNKNOWN, "\""+name+"\" unknown command or instruction");
			return false;
		}
	}
	else {
		Expected(ERR_EXP_ID, line);
		return false;
	}
}

// Fill out in output file label used before being defined
void UsedLabelFill(LabelUse *labelUse)
{
	int addr;
	if ( findLabel(labelUse->getCurrentPath(), labelUse->getLabelPath(), &addr) ) {
		if ( out::opened() ) {
			out::setPos( labelUse->getPos() );
			out::write(&addr, 4);
		}
	}
	else showError(ERR_COMPILER_USEDLBL, "\""+labelUse->getLabelPath()+"\" label used without being defined", labelUse->getFile(), labelUse->getNumber());
}

// Fill out in output file boot information from g_BootInfo
void FillBootInfo(bool NoDisk)
{
	if (out::opened()) {
		// set pos
		if (NoDisk) out::setPos(0);
		else out::setPos(out::HEADER_SIZE);
		// write boot info
		out::writeBootInfo(g_BootInfo);
	}
}

// Fill out in output file labels used before being defined
void UsedLabelsFill()
{
	g_usedLabels.setPosFirst();
	while ( !g_usedLabels.queryEnd() ) {
		UsedLabelFill( g_usedLabels.at() );
		g_usedLabels.setPosNext();
	}
}

void showStats()
{
	int outSize = -1; if ( out::opened() ) outSize = out::getSize();
	float time = max(0, GetTickCount()-stats::startTime) / 1000.0f;

	using namespace std;
	cout << "Instructions : " << stats::instructions << endl;
	cout << "Macros       : defined - " << stats::macroDefined << ", used - " << stats::macroUsed << endl;
	cout << "Labels       : defined - " << stats::labelDefined << ", used - " << stats::labelUsed << endl;
	cout << "Source       : lines - " << stats::lines << ", files - " <<  stats::files << ", size - " << stats::sourceSize << " (" << size2str(stats::sourceSize) << ')' << endl;
	if (outSize > -1) cout << "Output size  : " << outSize << " (" << size2str(outSize) << ')' << endl;
	cout << "Total time   : " << px::float2str(time, 'f', 2) << 's' << endl;
	cout << "Errors       : " << errorCount() << " (" << (errorCount() == 0 ? "SUCCESS" : "FAILURE") << "!)" << endl;
}

//-------------------------------------------------------------------------
// Public functions

bool compile(bool nodisk, bool nobootinfo)
{
	static const int DEFAULT_BOOT_INFO = -1;

	if (!LoadInstructions()) return false;

	if (errorCount() == 0) {
		out::open(g_FileDest);
		if (!nodisk)
			out::writeHeader();
		if (!nobootinfo) {
			g_BootInfo = DEFAULT_BOOT_INFO;
			out::writeBootInfo(g_BootInfo);
		}
	}

	g_source.setPosFirst();
	while (!g_source.queryEnd()) {
		compileLine();
		g_source.setPosNext();
	}

	// check namespace path
	if (!g_currentPath.empty())
		showError(ERR_COMPILER_NSEMPTY, "\""+g_currentPath+"\" unclosed namespace");

	// fill out output file
	if (!nobootinfo && g_BootInfo != DEFAULT_BOOT_INFO)
		FillBootInfo(nodisk);
	UsedLabelsFill();

	showStats();

	out::close();

	g_labels.Clear();
	g_usedLabels.Clear();

	return true;
}
