/*  NoASM compiler for NoSense - The Virtual Machine
    Copyright (C) 2004  Adam Sawicki "Regedit"

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//-------------------------------------------------------------------------
// Name    : Preprocessor
// Version : 1.0
// Desc    : Module with part 2/3 - preprocessor
// Author  : Adam Sawicki "Regedit"
// Date    : 2004-06-18
// URL     : mailto:regedit@risp.pl
// URL     : http://www.programex.prv.pl/
//-------------------------------------------------------------------------
#include "stdafx.h"
#include "base.h"
#include "Tokenizer.h"
#include "Preprocessor.h"

//-------------------------------------------------------------------------
// Private data types

class Macro : public TokenList, public NamedItem {
public:
	Macro(const std::string &a_name) : NamedItem(a_name) { }
};

typedef UniversalMap<Macro> MacroList;

//-------------------------------------------------------------------------
// Variables

MacroList g_macros;

//-------------------------------------------------------------------------
// Private functions

/* void debug_showMacros()
{
	std::cout << "[DEGUB] MACROS:" << std::endl;
	Macro *m;
	Token *t;
	std::string s;
	g_macros.setPosFirst();
	while (!g_macros.queryEnd()) {
		m = g_macros.at();
		std::cout << "- \"" << m->getName() << "\"\n";
		m->setPosFirst();
		while (!m->queryEnd()) {
			t = m->at();
			switch (t->type)
			{
			case Token::IDENTIFIER: s = "  identifier"; break;
			case Token::NUMBER: s = "  number"; break;
			case Token::STRING: s = "  string"; break;
			case Token::SYMBOL: s = "  symbol"; break;
			}
			s += " \"" + t->str + "\"";
			std::cout << s << std::endl;
			m->setPosNext();
		}
		g_macros.setPosNext();
	}
} */

/* Table of CCC changes according to:
- directive (a)
- current CCC value (b)
- condition truth (c)

         |          (b)          |
         |  CCC = 0  |  CCC > 0  |
         +-----------+-----------+
         |    (c)    |    (b)    |
(a)      |  T  |  F  | = 1 | > 1 |
---------+-----+-----+-----+-----+
if[n]def |  0    +1  |    +1     |
else     |    +1     | -1     0  |
endif    |     0     |    -1     | */

bool preprocess_ifdef(int *CCC)
{
	Line *line = g_source.at(); line->setPosFirst();
	std::string name;
	if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		if (strCiEq(name, "ifdef")) {
			std::string macroName;
			if (tokenParticular(line, Token::IDENTIFIER, &macroName)) {
				lineEnd(line);
				if (*CCC == 0) {
					if (g_macros.find(macroName) == 0) (*CCC)++;
				}
				else /* (*CCC > 0) */ (*CCC)++;
			} else Expected(ERR_EXP_ID, line);
			g_source.Delete(); return true;
		} else return false;
	} else return false;
}

bool preprocess_ifndef(int *CCC)
{
	Line *line = g_source.at(); line->setPosFirst();
	std::string name;
	if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		if (strCiEq(name, "ifndef")) {
			std::string macroName;
			if (tokenParticular(line, Token::IDENTIFIER, &macroName)) {
				lineEnd(line);
				if (*CCC == 0) {
					if (g_macros.find(macroName) != 0) (*CCC)++;
				}
				else /* (*CCC > 0) */ (*CCC)++;
			} else Expected(ERR_EXP_ID, line);
			g_source.Delete(); return true;
		} else return false;
	} else return false;
}

bool preprocess_else(int *CCC)
{
	Line *line = g_source.at(); line->setPosFirst();
	std::string name;
	if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		if (strCiEq(name, "else")) {
			lineEnd(line);
			if (*CCC == 0) (*CCC)++;
			else if (*CCC == 1) (*CCC)--;
			// (*CCC > 1) - no change
			g_source.Delete(); return true;
		} else return false;
	} else return false;
}

bool preprocess_endif(int *CCC)
{
	Line *line = g_source.at(); line->setPosFirst();
	std::string name;
	if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		if (strCiEq(name, "endif")) {
			lineEnd(line);
			if (*CCC > 0) (*CCC)--;
			// if (*CCC == 0) - no change
			g_source.Delete(); return true;
		} else return false;
	} else return false;
}

// Preprocess current source line as a "error" directive (if there isn't this directive - return false)
// Do all - read first identifier (directive name), check it, process and move line pointer in the source
bool preprocess_error()
{
	Line *line = g_source.at(); line->setPosFirst();
	std::string name;
	if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		if (strCiEq(name, "error")) {
			std::string msg;
			if (tokenParticular(line, Token::STRING, &msg)) {
				lineEnd(line);
				showError(ERR_USER, msg, line->getFile(), line->getNumber());
			} else Expected(ERR_EXP_STR, line);
			g_source.Delete(); return true;
		} else return false;
	} else return false;
}

bool preprocess_include()
{
	Line *line = g_source.at(); line->setPosFirst();
	std::string name;
	if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		if (strCiEq(name, "include")) {
			std::string fileName;
			if (tokenParticular(line, Token::STRING, &fileName)) {
				lineEnd(line);

				std::string err_file = line->getFile();
				int err_number = line->getNumber();

				LineList::iterator newPos;
				g_source.Delete();
				{
					std::string RealFileName = LookForFile(fileName);
					if (RealFileName.empty())
						showError(ERR_FILENOTFOUND, "\""+fileName+"\" file not found", err_file, err_number);
					else {
						if ( tokenize(RealFileName, &newPos, err_file, err_number) )
							g_source.setPos(newPos);
					}
				}
			} else { Expected(ERR_EXP_STR, line); g_source.Delete(); }
			return true;
		} else return false;
	} else return false;
}

bool preprocess_define()
{
	Line *line = g_source.at(); line->setPosFirst();
	std::string name;
	if (tokenParticular(line, Token::IDENTIFIER, &name)) {
		if (strCiEq(name, "define")) {
			std::string macroName;
			if (tokenParticular(line, Token::IDENTIFIER, &macroName)) {
				// if macro with this name already exists
				if (g_macros.find(macroName) != 0)
					showError(ERR_PRE_MACROEXISTS, "\""+macroName+"\" macro already exists", line->getFile(), line->getNumber());
				else {
					Macro *macro = new Macro(macroName);
					while (!line->queryEnd()) {
						macro->Add( new Token( *(line->at()) ) );
						line->setPosNext();
					}
					g_macros.Add(macro);
					stats::macroDefined++;
				}
			} else Expected(ERR_EXP_ID, line);
			g_source.Delete(); return true;
		} else return false;
	} else return false;
}

// Inserts tokens of given macro into given line at its current position
// Return in firstPos iterator of the first inserted token or NULL iterator if nothing inserted
void InsertMacro(Line *line, Macro *macro, TokenList::iterator *firstPos)
{
	*firstPos = line->getPosNULL();
	TokenList::iterator it;

	macro->setPosFirst();
	while (!macro->queryEnd())
	{
		it = line->Insert( new Token( *macro->at() ) );
		if ( *firstPos == line->getPosNULL() )
			*firstPos = it;
		macro->setPosNext();
	}
}

// Expand identifiers in the current source line from currently existing macros
void preprocess_expand()
{
	Line *l = g_source.at();
	Token *t;
	Macro *m;
	l->setPosFirst();
	while (!l->queryEnd()) {
		t = l->at();
		if (t->type == Token::IDENTIFIER) {
			m = g_macros.find(t->str);
			if (m != 0) {
				if (m->empty()) {
					showError(ERR_PRE_MACROEMPTY, "\""+m->getName()+"\" cannot expand empty macro", l->getFile(), l->getNumber());
					l->setPosNext();
				}
				else {
					// special macro
					if (m->getName() == "__FILE__") {
						m->setPosFirst();
						m->at()->str = g_source.at()->getFile();
					}
					else if (m->getName() == "__LINE__") {
						m->setPosFirst();
						m->at()->str = px::int2str(g_source.at()->getNumber());
					}

					TokenList::iterator newPos;
					l->Delete();
					InsertMacro(l, m, &newPos);
					l->setPos(newPos);
					stats::macroUsed++;
				}
			}
			else l->setPosNext();
		}
		else l->setPosNext();
	}
}

// Preprocesses current source line (on error: show message, return false)
bool preprocessLine(int *CCC)
{
	// if line is empty (it shound never be) - delete this line
	if (g_source.at()->empty()) {
		g_source.Delete();
		return true;
	}

	// conditional compilation counter > 0 - we are inside a block to skip
	if (*CCC > 0) {
		// process one of these directives
		if (preprocess_ifdef(CCC)) { }
		else if (preprocess_ifndef(CCC)) { }
		else if (preprocess_else(CCC)) { }
		else if (preprocess_endif(CCC)) { }
		// or delete this line
		else g_source.Delete();
		return true;
	}
	// conditional compilation counter = 0 - we are inside a block to leave
	else if (*CCC == 0) {
		// process any of implemented directives
		if (preprocess_ifdef(CCC)) { }
		else if (preprocess_ifndef(CCC)) { }
		else if (preprocess_else(CCC)) { }
		else if (preprocess_endif(CCC)) { }
		else if (preprocess_error()) { }
		else if (preprocess_include()) { }
		else if (preprocess_define()) { }
		// or if it is other line
		else {
			// expand identifiers from macros
			preprocess_expand();
			// leave this line - move to next
			g_source.setPosNext();
		}
		return true;
	}
	// this should never be
	else {
		showError(ERR_INTERNAL, "Preprocessor internal error: " + px::int2str(*CCC) + " conditional compilation counter cannot be negative", g_source.at()->getFile(), g_source.at()->getNumber());
		return false;
	}
}

void AddPredefinedMacros()
{
	Macro *m;

	std::string date, time;
	getDateTime(&date, &time);

	// __FILE__
	m = new Macro("__FILE__");
	m->Add( new Token(Token::STRING, "ERROR") );
	g_macros.Add(m);

	// __LINE__
	m = new Macro("__LINE__");
	m->Add( new Token(Token::NUMBER, "-1") );
	g_macros.Add(m);

	// __DATE__
	m = new Macro("__DATE__");
	m->Add( new Token(Token::STRING, date) );
	g_macros.Add(m);

	// __TIME__
	m = new Macro("__TIME__");
	m->Add( new Token(Token::STRING, time) );
	g_macros.Add(m);

	// __VERSION__
	m = new Macro("__VERSION__");
	m->Add( new Token(Token::STRING, VERSION_STR) );
	g_macros.Add(m);
}

//-------------------------------------------------------------------------
// Public functions

// Preprocess source (on error: show message, return false)
bool preprocess()
{
	// before preprocessing add predefined macros
	AddPredefinedMacros();

	// conditional compilation counter
	// = 0 - let the line go
	// > 0 - delete the line
	int CCC = 0;
	bool ok = true;
	g_source.setPosFirst();
	while (!g_source.queryEnd())
		if (!preprocessLine(&CCC)) {
			ok = false;
			break;
		}
//	debug_showMacros();

	// after preprocessing macro list can be cleared
	g_macros.Clear();
	return ok;
}
