/*  NoASM compiler for NoSense - The Virtual Machine
    Copyright (C) 2004  Adam Sawicki "Regedit"

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//-------------------------------------------------------------------------
// Name    : Tokenizer
// Version : 1.0
// Desc    : Module with part 1/3 - tokenizer
// Author  : Adam Sawicki "Regedit"
// Date    : 2004-06-18
// URL     : mailto:regedit@risp.pl
// URL     : http://www.programex.prv.pl/
//-------------------------------------------------------------------------
#include "stdafx.h"
#include "base.h"
#include "Tokenizer.h"

//-------------------------------------------------------------------------
// Parsing functions

// success : index is moved, return true
// failure : index is unchanged, return false
// Every parsing function must check range of the position.

std::string parsingFile;
int parsingLine;

bool parseLetter(const std::string &s, size_t *index, char *c)
{
	if ( *index >= s.size() ) return false;
	*c = s[*index];
	if ( (*c >= 'a' && *c <= 'z') || (*c >= 'A' && *c <= 'Z') || (*c == '_') ) {
		(*index)++;
		return true;
	}
	else return false;
}

bool parseDigit(const std::string &s, size_t *index, char *c)
{
	if ( *index >= s.size() ) return false;
	*c = s[*index];
	if (*c >= '0' && *c <= '9') {
		(*index)++;
		return true;
	}
	else return false;
}

bool parseParticularChar(const std::string &s, size_t *index, char ch, char *c)
{
	if ( *index >= s.size() ) return false;
	if (s[*index] != ch) return false;
	if (c) *c = ch;
	(*index)++;
	return true;
}

// One or more whitespaces (except end of line characters)
bool parseWhitespace(const std::string &s, size_t *index)
{
	bool ok = false;
	while ( *index < s.size() && (s[*index] == ' ' || s[*index] == '\t') ) {
		(*index)++;
		ok = true;
	}
	return ok;
}

// Comment that begins with ';' and ends with end of the string
bool parseComment(const std::string &s, size_t *index)
{
	if ( *index >= s.size() ) return false;
	if ( s[*index] != ';' ) return false;
	*index = s.size();
	return true;
}

// Any correct identifier name
bool parseIdentifier(const std::string &s, size_t *index, std::string *name)
{
	char c;
	name->clear();
	if ( parseLetter(s, index, &c) ) {
		*name += c;
		for (;;) {
			if ( parseLetter(s, index, &c) ) {
				*name += c;
			}
			else if ( parseDigit(s, index, &c) ) {
				*name += c;
			}
			else break;
		}
		return true;
	}
	else return false;
}

bool parseSymbol(const std::string &s, size_t *index, std::string *symbol)
{
	if (*index >= s.size()) return false;
	if (s[*index] == ',' || s[*index] == ':' || s[*index] == '&' || s[*index] == '^' || s[*index] == '*' || s[*index] == '.') {
		*symbol = s[*index];
		(*index)++;
		return true;
	}
	else return false;
}

bool parseNumber(const std::string &s, size_t *index, std::string *number)
{
	char c;
	number->clear();
	if ( parseDigit(s, index, &c) || parseParticularChar(s, index, '+', &c) || parseParticularChar(s, index, '-', &c) ) {
		*number += c;
		for (;;) {
			if ( parseDigit(s, index, &c) ) *number += c;
			else if ( parseLetter(s, index, &c) ) *number += c;
			else if ( parseParticularChar(s, index, '+', &c) ) *number += c;
			else if ( parseParticularChar(s, index, '-', &c) ) *number += c;
			else if ( parseParticularChar(s, index, '.', &c) ) *number += c;
			else break;
		}
		return true;
	}
	else return false;
}

bool parseEscapedCharacterNumeric(const std::string &s, size_t *index, char *c)
{
	size_t index_bak = *index;
	if (*index >= s.size()) return false;
	std::string str;
	char ch;
	bool hex = false;
	if (s[*index] == 'x' || s[*index] == 'X') {
		(*index)++;
		hex = true;
	}
	while (*index < s.size()) {
		ch = s[*index];
		if ( (ch >= '0' && ch <= '9') || (hex && ((ch >= 'a' && ch <= 'f') || (ch >= 'A' && ch <= 'F'))) ) {
			str += ch;
			(*index)++;
		}
		else break;
	}
	if (str.empty()) {
		*index = index_bak;
		return false;
	}
	px::LowerCase(&str);
	int i;
	if (hex) i = px::hex2int(str);
	else i = px::str2int(str);
	if (i < 0 || i >= 256) {
		*index = index_bak;
		return false;
	}
	else {
		*c = static_cast<char>(i);
		return true;
	}
}

bool parseEscapedCharacter(const std::string &s, size_t *index, char *c)
{
	if (*index >= s.size()) return false;
	if (s[*index] == '\\') {
		*c = '\\';
		(*index)++;
		if (*index >= s.size()) {
			showError(ERR_TOKENIZERESC1, "Syntax error in escape character", parsingFile, parsingLine);
			return true;
		}
		else if (s[*index] == 'b') { *c = '\b'; (*index)++; return true; }
		else if (s[*index] == 'f') { *c = '\f'; (*index)++; return true; }
		else if (s[*index] == 'n') { *c = '\n'; (*index)++; return true; }
		else if (s[*index] == 'r') { *c = '\r'; (*index)++; return true; }
		else if (s[*index] == 't') { *c = '\t'; (*index)++; return true; }
		else if (s[*index] == 'v') { *c = '\v'; (*index)++; return true; }
		else if (s[*index] == 'a') { *c = '\a'; (*index)++; return true; }
		else if (s[*index] == '\\') { *c = '\\'; (*index)++; return true; }
		else if (s[*index] == '\'') { *c = '\''; (*index)++; return true; }
		else if (s[*index] == '"') { *c = '"'; (*index)++; return true; }
		else if (s[*index] == '0') { *c = '\0'; (*index)++; return true; }
		else if (s[*index] == '?') { *c = '?'; (*index)++; return true; }
		else if (parseEscapedCharacterNumeric(s, index, c)) return true;
		else {
			showError(ERR_TOKENIZERESC2, "Unknown escape character", parsingFile, parsingLine);
			return true;
		}
	}
	else return false;
}

bool parseString(const std::string &s, size_t *index, std::string *str)
{
	char c;
	if (*index >= s.size()) return false;
	char delimiter = s[*index];
	if (delimiter == '\'' || delimiter == '"') {
		(*index)++;
		str->clear();
		for(;;) {
			if (*index >= s.size()) {
				showError(ERR_TOKENIZERSTR1, "Unexpected end of string", parsingFile, parsingLine);
				return true;
			}
			else if (parseEscapedCharacter(s, index, &c)) {
				*str += c;
			}
			// end of the string
			else if (s[*index] == delimiter) {
				(*index)++;
				return true;
			}
			// other character
			else {
				*str += s[*index];
				(*index)++;
			}
		}
	}
	else return false;
}

//-------------------------------------------------------------------------
// Other functions

// - Tokenize given line
// - Add tokenized line into g_source at the current position
// - In newLine return iterator to inserted line or NULL pos if nothing inserted
// (on error: show message, return false)
bool tokenizeLine(const std::string &line, const std::string &fileName, int number, LineList::iterator *newLine)
{
	parsingFile = fileName;
	parsingLine = number;
	bool ok = true;
// DEBUG
//	std::cout << fileName << ' ' << number << " \"" << line << '"' << std::endl;
	Line *l = new Line(fileName, number);
	// Read tokens
	size_t index = 0;
	parseWhitespace(line, &index);
	std::string s;
	while (index < line.size()) {
		if (parseSymbol(line, &index, &s))
			l->Add( new Token(Token::SYMBOL, s) );
		else if (parseIdentifier(line, &index, &s))
			l->Add( new Token(Token::IDENTIFIER, s) );
		else if (parseNumber(line, &index, &s) )
			l->Add( new Token(Token::NUMBER, s) );
		else if (parseString(line, &index, &s) )
			l->Add( new Token(Token::STRING, s) );
		else if (parseComment(line, &index)) { }
		else {
			showError(ERR_TOKENIZER, "General syntax error", fileName, number);
			ok = (l->getCount() > 0);
			break;
		}
		parseWhitespace(line, &index);
	}
	if (ok) {
		if (l->getCount() == 0) {
			*newLine = g_source.getPosNULL();
			delete l;
		}
		else {
			*newLine = g_source.Insert(l);
			stats::lines++;
		}
	}
	else delete l;
	return ok;
}

//-------------------------------------------------------------------------
// Public functions

// - Load given source code file
// - Split it into lines
// - Tokenize this lines
// - Add tokenized lines into g_source at the current position
// - In firstLine return iterator to the first inserted line or NULL position if nothing inserted
// (on error: show message, return false)
bool tokenize(const std::string &FileName, LineList::iterator *firstLine, const std::string &incFile, int incLine)
{
	if (firstLine) *firstLine = g_source.getPosNULL();
	// Load file to string
	std::string loadedFile;
	if ( !px::LoadStringFromFile(FileName, loadedFile) ) {
		if (!incFile.empty())
			showError(ERR_SRCLOAD, "Cannot load source file \""+FileName+"\"", incFile, incLine);
		else
			showError(ERR_SRCLOAD, "Cannot load source file \""+FileName+"\"");
		return false;
	}
	stats::sourceSize += static_cast<int>( loadedFile.size() );
	// Read lines
	LineList::iterator newLine;
	std::string line;
	size_t index = 0;
	int number = 1;
	bool was13 = false;
	while ( index < loadedFile.size() ) {
		if (loadedFile[index] == '\r' || loadedFile[index] == '\n') {
			if (!line.empty()) {
				if (!tokenizeLine(line, px::ExtractFileName(FileName), number, &newLine)) return false;
				if (firstLine) if (*firstLine == g_source.getPosNULL()) *firstLine = newLine;
				line.clear();
			}
			if ( !(was13 && loadedFile[index] == '\n') ) number++;
		}
		else line += loadedFile[index];
		was13 = (loadedFile[index] == '\r');
		index++;
	}
	if (!line.empty()) {
		if (!tokenizeLine(line, px::ExtractFileName(FileName), number, &newLine)) return false;
		if (firstLine) if (*firstLine == g_source.getPosNULL()) *firstLine = newLine;
	}
	stats::files++;
	return true;
}
