/*  NoASM compiler for NoSense - The Virtual Machine
    Copyright (C) 2004  Adam Sawicki "Regedit"

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//-------------------------------------------------------------------------
// Name    : Base
// Version : 1.0
// Desc    : Module with general common routines
// Author  : Adam Sawicki "Regedit"
// Date    : 2004-06-13
// URL     : mailto:regedit@risp.pl
// URL     : http://www.programex.prv.pl/
//-------------------------------------------------------------------------
#include "stdafx.h"
#include <pxxnl.h>
#include "output.h"
#include "base.h"

//-------------------------------------------------------------------------
// General functions

bool strCiEq(const std::string &s1, const std::string &s2)
{
	return (px::LowerCase(s1) == px::LowerCase(s2));
}

void debug_showSource()
{
	std::cout << "[DEGUB] SOURCE LINES:" << std::endl;
	Line *l;
	Token *t;
	std::string s;
	g_source.setPosFirst();
	while (!g_source.queryEnd()) {
		l = g_source.at();
		std::cout << "-";
		l->setPosFirst();
		while (!l->queryEnd()) {
			t = l->at();
			switch (t->type)
			{
			case Token::IDENTIFIER: s = " id"; break;
			case Token::NUMBER: s = " nr"; break;
			case Token::STRING: s = " str"; break;
			case Token::SYMBOL: s = " sym"; break;
			}
			s += " \"" + t->str + "\"";
			std::cout << s;
			l->setPosNext();
		}
		std::cout << std::endl;
		g_source.setPosNext();
	}
}

std::string word2str(WORD w, size_t minDigits)
{
	std::string s = px::int2str(static_cast<int>(w));
	while (s.size() < minDigits) s = '0'+s;
	return s;

}

void getDateTime(std::string *date, std::string *time)
{
	SYSTEMTIME st;
	GetSystemTime(&st);
	*date = word2str(st.wYear) + "-" + word2str(st.wMonth, 2) + "-" + word2str(st.wDay, 2);
	*time = word2str(st.wHour, 2) + ":" + word2str(st.wMinute, 2) + ":" + word2str(st.wSecond, 2);
}

void generateRandomString(std::string *s)
{
	*s = '_';
	*s += px::int2str(rand());
	*s += '_';
	*s += px::int2str(rand());
}

std::string size2str(int size)
{
	if (size >= 1024*1024*1024)
		return px::int2str( size/(1024*1024*1024) ) + "GB";
	if (size >= 1024*1024)
		return px::int2str( size/(1024*1024) ) + "MB";
	if (size >= 1024)
		return px::int2str( size/(1024) ) + "kB";
	return px::int2str(size) + "B";
}

//-------------------------------------------------------------------------
// Erroring

const char *ERRMSG_DESTWRITE              = "Cannot write to output file";
const char *ERRMSG_COMPILER_ADDRMODE1     = "Only addressing mode 1 allowed here";
const char *ERRMSG_COMPILER_NOLABEL       = "Label not allowed here";
const char *ERRMSG_COMPILER_BOOTINFOREDEF = "Boot info already defined";


const int ERROR_MSG_LIMIT = 8;

int g_errorCount = 0;

void showError(int code, const std::string &message)
{
	out::cancel();
	g_errorCount++;
	if (g_errorCount < ERROR_MSG_LIMIT)
		std::cout << "Error " << code << ": " << message << std::endl;
	else if (g_errorCount == ERROR_MSG_LIMIT)
		std::cout << "and more errors..." << std::endl;
}

void showError(int code, const std::string &message, const std::string &file, int line)
{
	out::cancel();
	g_errorCount++;
	if (g_errorCount < ERROR_MSG_LIMIT)
		std::cout << file << " (" << line << "): error " << code << ": " << message << std::endl;
	else if (g_errorCount == ERROR_MSG_LIMIT)
		std::cout << "and more errors..." << std::endl;
}

int errorCount() { return g_errorCount; }

//-------------------------------------------------------------------------
// OptimizedStringDatabase class

OPTIMIZED_STRING_MAP::iterator OptimizedStringDatabase::Add(const std::string &s)
{
	// search element with given key
	OPTIMIZED_STRING_MAP::iterator it = map.find(s);
	// if not found
	if ( it == End() )
		// then add it and return it
		return map.insert( std::make_pair(s, 1) ).first;
	// and if found
	else {
		// increment reference counter
		it->second++;
		// and return it
		return it;
	}
}

void OptimizedStringDatabase::Delete(OPTIMIZED_STRING_MAP::iterator it)
{
	// decrement reference counter
	it->second--;
	// if it dropped to zero
	if (it->second == 0)
		// then delete it
		map.erase(it);
}

//-------------------------------------------------------------------------
// OptimizedString class

OptimizedString::OptimizedString()
{
	m_it = g_OptimizedStringDatabase.End();
}

OptimizedString::OptimizedString(const std::string &s)
{
	m_it = g_OptimizedStringDatabase.Add(s);
}

OptimizedString::OptimizedString(const OptimizedString &Source)
{
	m_it = g_OptimizedStringDatabase.Add( Source.Get() );
}

OptimizedString::~OptimizedString()
{
	if ( m_it != g_OptimizedStringDatabase.End() )
		g_OptimizedStringDatabase.Delete(m_it);
}

void OptimizedString::Set(const std::string &s)
{
	// if old string isn't empty
	if ( m_it != g_OptimizedStringDatabase.End() )
		// then delete it
		g_OptimizedStringDatabase.Delete(m_it);

	// if new string isn't empty
	if ( !s.empty() )
		// then create it
		m_it = g_OptimizedStringDatabase.Add(s);
	// and if new string is empty
	else
		// then NULL :)
		m_it = g_OptimizedStringDatabase.End();
}

std::string OptimizedString::Get() const
{
	if ( m_it == g_OptimizedStringDatabase.End() )
		return std::string();
	else
		return m_it->first;
}

//-------------------------------------------------------------------------
// Token reading functions

bool tokenAny(Line *l, Token::TYPE *type, std::string *str)
{
	if (l->queryEnd()) return false;
	else {
		if (type) *type = l->at()->type;
		if (str) *str = l->at()->str;
		l->setPosNext();
		return true;
	}
}

bool tokenParticular(Line *l, Token::TYPE type, std::string *str)
{
	if (l->queryEnd()) return false;
	else {
		if (l->at()->type == type) {
			if (str) *str = l->at()->str;
			l->setPosNext();
			return true;
		}
		else return false;
	}
}

bool tokenParticular(Line *l, Token::TYPE type, const std::string &str)
{
	if (l->queryEnd()) return false;
	else {
		if (l->at()->type == type && l->at()->str == str) {
			l->setPosNext();
			return true;
		}
		else return false;
	}
}

void Expected(int type, Line *line)
{
	switch (type)
	{
	case ERR_EXP_STR: showError(type, "String expected", line->getFile(), line->getNumber()); break;
	case ERR_EXP_ID: showError(type, "Identifier expected", line->getFile(), line->getNumber()); break;
	case ERR_EXP_EOL: showError(type, "End of line expected", line->getFile(), line->getNumber()); break;
	case ERR_EXP_PARAM: showError(type, "Parameter expected", line->getFile(), line->getNumber()); break;
	case ERR_EXP_NR: showError(type, "Number expected", line->getFile(), line->getNumber()); break;
	default: showError(ERR_INTERNAL, "Internal error in Expected()", line->getFile(), line->getNumber());
	}
}

bool lineEnd(Line *l)
{
	if (l->queryEnd()) return true;
	else {
		Expected(ERR_EXP_EOL, l);
		return false;
	}
}

//-------------------------------------------------------------------------
// Global variables

OptimizedStringDatabase g_OptimizedStringDatabase;
LineList g_source;

int stats::files = 0;
int stats::instructions = 0;
int stats::lines = 0;
int stats::sourceSize = 0;
DWORD stats::startTime;
int stats::macroDefined = 0;
int stats::macroUsed = 0;
int stats::labelDefined = 0;
int stats::labelUsed = 0;

//-------------------------------------------------------------------------
// Include directories list

StringList g_IncDir;
std::string g_FileSrc;
std::string g_FileDest;
std::string g_PathSrc;

bool LoadIncDir()
{
	std::string docFileName = "Directories.xnl";
	std::string doc_s, dir;
	bool ok = px::LoadStringFromFile(px::GetModulePath()+docFileName, doc_s);

	if (ok) {
		px::CXNLRead doc(&doc_s);
		doc.Begin(px::GetModulePath());
		while (!doc.QueryEOF()) {
			if (doc.ReadElement(&dir))
				g_IncDir.push_back(dir);
			else {
				ok = false;
				break;
			}
		}
		doc.End();
	}

	if (!ok) showError(ERR_INCDIRS, "Cannot load include directories from file \""+docFileName+"\"");
	return ok;
}

std::string LookForFile(const std::string &FileName)
{
	// source file directory
	std::string s = px::Relative2AbsolutePath(g_PathSrc, FileName);
	if (px::FileExists(s)) return s;
	// include directories
	for (StringList::iterator it = g_IncDir.begin(); it != g_IncDir.end(); ++it) {
		s = px::Relative2AbsolutePath(*it, FileName);
		if (px::FileExists(s)) return s;
		s = px::Absolute2RelativePath(px::GetModulePath(), px::Relative2AbsolutePath(*it, FileName));
		if (px::FileExists(s)) return s;
	}
	// not found :(
	return std::string();
}