/*  NoASM compiler for NoSense - The Virtual Machine
    Copyright (C) 2004  Adam Sawicki "Regedit"

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//-------------------------------------------------------------------------
// Name    : Base
// Version : 1.0
// Desc    : Module with general common routines
// Author  : Adam Sawicki "Regedit"
// Date    : 2004-06-13
// URL     : mailto:regedit@risp.pl
// URL     : http://www.programex.prv.pl/
//-------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Constants

const std::string VERSION_STR = "1.0.1";

//-------------------------------------------------------------------------
// General functions

// Compare two strings without case sensitivity
bool strCiEq(const std::string &s1, const std::string &s2);
// Show source - lines and they tokens
void debug_showSource();
// Convert WORD to std::string filling up to minDigits characters with zeros
std::string word2str(WORD w, size_t minDigits = 0);
// Return strings with current date and time
void getDateTime(std::string *date, std::string *time);
// Generates unique string with random characters (that is valid identifier)
void generateRandomString(std::string *s);
// Convert size in bytes to string like "1 kB"
std::string size2str(int size);
//-------------------------------------------------------------------------
// Erroring

// Every error has integer code and string message
// Not all the error messages will be shown - only up to fixed count limit

const int ERR_SRCLOAD                = 100;
const int ERR_DESTOPEN               = 200;
const int ERR_DESTWRITEHEADER        = 201;
const int ERR_DESTWRITEBOOTINFO      = 202;
const int ERR_DESTWRITE              = 203;
const int ERR_INSTRUCTIONS           = 300;
const int ERR_TOKENIZER              = 400;
const int ERR_TOKENIZERSTR1          = 401;
const int ERR_TOKENIZERESC1          = 402;
const int ERR_TOKENIZERESC2          = 403;
const int ERR_EXP_STR                = 500;
const int ERR_EXP_ID                 = 501;
const int ERR_EXP_EOL                = 502;
const int ERR_EXP_PARAM              = 503;
const int ERR_EXP_NR                 = 504;
const int ERR_USER                   = 600;
const int ERR_INTERNAL               = 700;
const int ERR_PRE_MACROEXISTS        = 800;
const int ERR_PRE_MACROEMPTY         = 801;
const int ERR_COMPILER_UNKNOWN       = 900;
const int ERR_COMPILER_XTYPE         = 901;
const int ERR_COMPILER_NSEND         = 902;
const int ERR_COMPILER_NSEMPTY       = 903;
const int ERR_COMPILER_LBLREDEF      = 904;
const int ERR_COMPILER_USEDLBL       = 905;
const int ERR_COMPILER_ADDRMODE1     = 906;
const int ERR_COMPILER_NOLABEL       = 907;
const int ERR_COMPILER_INCLUDEBIN    = 908;
const int ERR_COMPILER_BOOTINFOREDEF = 909;
const int ERR_INCDIRS                = 1000;
const int ERR_FILENOTFOUND           = 1001;

extern const char *ERRMSG_COMPILER_ADDRMODE1;
extern const char *ERRMSG_COMPILER_NOLABEL;
extern const char *ERRMSG_COMPILER_BOOTINFOREDEF;
extern const char *ERRMSG_DESTWRITE;

void showError(int code, const std::string &message);
void showError(int code, const std::string &message, const std::string &file, int line);
int errorCount();

//-------------------------------------------------------------------------
// Optimized string system

// This database (implemented on map) is useful to keep many same
// string objects - it remembers that string only once.

typedef std::map<std::string, size_t> OPTIMIZED_STRING_MAP;
typedef std::pair<std::string, size_t> OPTIMIZED_STRING_PAIR;

class OptimizedStringDatabase {
private:
	OPTIMIZED_STRING_MAP map;
public:
	size_t Count() { return map.size(); }
	OPTIMIZED_STRING_MAP::iterator End() { return map.end(); }

	OPTIMIZED_STRING_MAP::iterator Add(const std::string &s);
	void Delete(OPTIMIZED_STRING_MAP::iterator it);
};

// Externally it works as std::string (maybe except a bit more difficult usage)
// It means - contains character string.
// Internally uses optimization containing the same string in many objects only once.
class OptimizedString
{
private:
	OPTIMIZED_STRING_MAP::iterator m_it;
public:
	OptimizedString(); // Default constructor
	OptimizedString(const std::string &s); // Initializing constructor
	OptimizedString(const OptimizedString &Source); // Copy constructor
	~OptimizedString(); // Destructor
	void Set(const std::string &s);
	std::string Get() const;
};

//-------------------------------------------------------------------------
// Data structures

typedef std::list<std::string> StringList;

template<typename T>
class UniversalList {
protected:
	typedef std::list<T*> ListType;
public:
	typedef typename ListType::iterator iterator;
protected:
	ListType list;
	iterator pos;
public:
	UniversalList() { pos = list.begin(); }
	// Delete all elements in the list
	virtual ~UniversalList();
	// Return true if the list is empty (no elements inside)
	bool empty() { return (list.size() == 0); }
	// Return element count
	size_t getCount() { return list.size(); }

	// Set pointer to the first element
	void setPosFirst() { pos = list.begin(); }
	// Set pointer to the next element
	void setPosNext() { ++pos; }
	// Set pointer to the previous element
	void setPosPrev() { --pos; }
	// Return true if pointer is at the end of the list
	bool queryEnd() { return (pos == list.end()); }
	// Return NULL pointer
	iterator getPosNULL() { return list.end(); }
	// Return current pointer
	iterator getPos() { return pos; }
	// Change current pointer
	void setPos(iterator it) { pos = it; }
	// Return object from current pointer
	T *at() { return (*pos); }

	// Add new element at the end of the list (element must be already created)
	void Add(T *x) { list.push_back(x); }
	// Add new element at current position and increment this position (element must be already created)
	iterator Insert(T *x);
	// Delete current element
	void Delete() { delete (*pos); pos = list.erase(pos); }
	// Delete all elements
	void Clear();

};

template<typename T>
UniversalList<T>::~UniversalList()
{
	Clear();
}

template<typename T>
typename UniversalList<T>::iterator UniversalList<T>::Insert(T *x)
{
	iterator r = list.insert(pos, x);
	pos = r;
	++pos;
	return r;
}

template<typename T>
void UniversalList<T>::Clear()
{
	for (iterator it = list.begin(); it != list.end(); ++it)
		delete (*it);
	list.clear();
}

// Abstract base class for all classes that has a name
class NamedItem {
protected:
	std::string name;
public:
	NamedItem() { }
	NamedItem(const std::string a_name) : name(a_name) { }
	const std::string &getName() const { return name; }
};

// Collection of named items contained in a map remembering lower case
// names of items. It has case-insensitive search.
template <typename T>
class UniversalMap {
private:
	// First element of this pair constains lowercase name
	typedef std::pair<std::string, T*> PairType;
	typedef std::map<std::string, T*> MapType;
	MapType map;
public:
	~UniversalMap();
	void Add(T *x);
	// Return item with given name (case insensitive comparison) or 0 if not found
	T *find(const std::string &name);
	void Clear();
};

template <typename T>
UniversalMap<T>::~UniversalMap()
{
	Clear();
}

template <typename T>
void UniversalMap<T>::Add(T *x)
{
	map.insert(make_pair( px::LowerCase(x->getName()), x ));
}

template <typename T>
T *UniversalMap<T>::find(const std::string &name)
{
	MapType::iterator it = map.find( px::LowerCase(name) );
	if (it == map.end()) return 0;
	else return it->second;
}

template <typename T>
void UniversalMap<T>::Clear()
{
	for (MapType::iterator it = map.begin(); it != map.end(); ++it)
		delete it->second;
	map.clear();
}

//-------------------------------------------------------------------------
// Particular data types

struct Token {
	enum TYPE { IDENTIFIER, SYMBOL, STRING, NUMBER };
	TYPE type;
	std::string str;
	Token(TYPE a_type, const std::string &a_str) : type(a_type), str(a_str) { }
	Token(const Token &t) : type(t.type), str(t.str) { }
};

typedef UniversalList<Token> TokenList;

class Line : public TokenList {
private:
	int number;
	OptimizedString file;
public:
	Line(const std::string &a_file, int a_number) : number(a_number), file(a_file) { }
	std::string getFile() { return file.Get(); }
	int getNumber() { return number; }
};

typedef UniversalList<Line> LineList;

//-------------------------------------------------------------------------
// Token reading functions

// Returns information from next token in given line (on error: return false)
// Moves token pointer in this line forward
bool tokenAny(Line *l, Token::TYPE *type, std::string *str);
bool tokenParticular(Line *l, Token::TYPE type, std::string *str);
// Succeedes only if the token has given string
bool tokenParticular(Line *l, Token::TYPE type, const std::string &str);
// Show error message telling that particular data type was expected
// (type must be one of ERR_EXP_ constants)
void Expected(int type, Line *line);
// If token pointer in given line isn't at the end, show message that
// end of line is expected and return false
bool lineEnd(Line *l);

//-------------------------------------------------------------------------
// Main common global variables

extern OptimizedStringDatabase g_OptimizedStringDatabase;
extern LineList g_source;

namespace stats {
	extern int files;
	extern int instructions;
	extern int lines;
	extern int sourceSize;
	extern DWORD startTime;
	extern int macroDefined;
	extern int macroUsed;
	extern int labelDefined;
	extern int labelUsed;
}

//-------------------------------------------------------------------------
// Include directories list

extern StringList g_IncDir;
// name of main source file and destination file
extern std::string g_FileSrc;
extern std::string g_FileDest;
extern std::string g_PathSrc;

// Load list of include directories from compiler configuration file
bool LoadIncDir();
// Return name of file to include with full path
// Given file name can be absolute or relative
// Functions does search according to main source file directory and
//   include directorires and checks where file exists
// (on error: return "")
std::string LookForFile(const std::string &FileName);
